@@ -1,5 +1,8 @@
 /*
- * Copyright (C) 2011 Karel Zak <kzak@redhat.com>
+ * No copyright is claimed.  This code is in the public domain; do with
+ * it what you wish.
+ *
+ * Written by Karel Zak <kzak@redhat.com>
  *
  * -- based on mount/losetup.c
  *
@@ -37,6 +40,8 @@
 #include "pathnames.h"
 #include "loopdev.h"
 #include "canonicalize.h"
+#include "at.h"
+#include "blkdev.h"
 
 #define CONFIG_LOOPDEV_DEBUG
 
@@ -64,8 +69,12 @@ loopdev_debug(const char *mesg, ...)
 # define DBG(m,x) do { ; } while(0)
 #endif
 
-
+/*
+ * see loopcxt_init()
+ */
 #define loopcxt_ioctl_enabled(_lc)	(!((_lc)->flags & LOOPDEV_FL_NOIOCTL))
+#define loopcxt_sysfs_available(_lc)	(!((_lc)->flags & LOOPDEV_FL_NOSYSFS)) \
+					 && !loopcxt_ioctl_enabled(_lc)
 
 /*
  * @lc: context
@@ -82,8 +91,10 @@ int loopcxt_set_device(struct loopdev_cxt *lc, const char *device)
 	if (!lc)
 		return -EINVAL;
 
-	if (lc->fd >= 0)
+	if (lc->fd >= 0) {
 		close(lc->fd);
+		DBG(lc, loopdev_debug("closing old open fd"));
+	}
 	lc->fd = -1;
 	lc->mode = 0;
 	lc->has_info = 0;
@@ -143,24 +154,42 @@ int loopcxt_has_device(struct loopdev_cxt *lc)
  */
 int loopcxt_init(struct loopdev_cxt *lc, int flags)
 {
+	int rc;
 	struct stat st;
+	struct loopdev_cxt dummy = UL_LOOPDEVCXT_EMPTY;
 
 	if (!lc)
 		return -EINVAL;
 
-	memset(lc, 0, sizeof(*lc));
+	memcpy(lc, &dummy, sizeof(dummy));
 	lc->flags = flags;
-	loopcxt_set_device(lc, NULL);
+
+	if (getenv("LOOPDEV_DEBUG"))
+		loopcxt_enable_debug(lc, TRUE);
+
+	rc = loopcxt_set_device(lc, NULL);
+	if (rc)
+		return rc;
+
+	if (stat(_PATH_SYS_BLOCK, &st) || !S_ISDIR(st.st_mode)) {
+		lc->flags |= LOOPDEV_FL_NOSYSFS;
+		lc->flags &= ~LOOPDEV_FL_NOIOCTL;
+		DBG(lc, loopdev_debug("init: disable /sys usage"));
+	}
 
 	if (!(lc->flags & LOOPDEV_FL_NOSYSFS) &&
-	    get_linux_version() >= KERNEL_VERSION(2,6,37))
+	    get_linux_version() >= KERNEL_VERSION(2,6,37)) {
 		/*
 		 * Use only sysfs for basic information about loop devices
 		 */
 		lc->flags |= LOOPDEV_FL_NOIOCTL;
+		DBG(lc, loopdev_debug("init: ignore ioctls"));
+	}
 
-	if (!(lc->flags & LOOPDEV_FL_CONTROL) && !stat(_PATH_DEV_LOOPCTL, &st))
+	if (!(lc->flags & LOOPDEV_FL_CONTROL) && !stat(_PATH_DEV_LOOPCTL, &st)) {
 		lc->flags |= LOOPDEV_FL_CONTROL;
+		DBG(lc, loopdev_debug("init: loop-control detected "));
+	}
 
 	return 0;
 }
@@ -172,6 +201,8 @@ int loopcxt_init(struct loopdev_cxt *lc, int flags)
  */
 void loopcxt_deinit(struct loopdev_cxt *lc)
 {
+	int errsv = errno;
+
 	if (!lc)
 		return;
 
@@ -180,8 +211,10 @@ void loopcxt_deinit(struct loopdev_cxt *lc)
 	free(lc->filename);
 	lc->filename = NULL;
 
-	loopcxt_set_device(lc, NULL);
+	ignore_result( loopcxt_set_device(lc, NULL) );
 	loopcxt_deinit_iterator(lc);
+
+	errno = errsv;
 }
 
 /*
@@ -203,7 +236,7 @@ void loopcxt_enable_debug(struct loopdev_cxt *lc, int enable)
  */
 char *loopcxt_strdup_device(struct loopdev_cxt *lc)
 {
-	if (!lc || !*lc->device)
+	if (!lc || !lc->device || !*lc->device)
 		return NULL;
 	return strdup(lc->device);
 }
@@ -215,7 +248,7 @@ char *loopcxt_strdup_device(struct loopdev_cxt *lc)
  */
 const char *loopcxt_get_device(struct loopdev_cxt *lc)
 {
-	return lc ? lc->device : NULL;
+	return lc && lc->device && *lc->device ? lc->device : NULL;
 }
 
 /*
@@ -257,8 +290,10 @@ int loopcxt_get_fd(struct loopdev_cxt *lc)
 
 	if (lc->fd < 0) {
 		lc->mode = lc->flags & LOOPDEV_FL_RDWR ? O_RDWR : O_RDONLY;
-		lc->fd = open(lc->device, lc->mode);
-		DBG(lc, loopdev_debug("open %s", lc->fd < 0 ? "failed" : "ok"));
+		lc->fd = open(lc->device, lc->mode | O_CLOEXEC);
+		DBG(lc, loopdev_debug("open %s [%s]: %s", lc->device,
+				lc->flags & LOOPDEV_FL_RDWR ? "rw" : "ro",
+				lc->fd < 0 ? "failed" : "ok"));
 	}
 	return lc->fd;
 }
@@ -332,8 +367,11 @@ int loopcxt_deinit_iterator(struct loopdev_cxt *lc)
 	free(iter->minors);
 	if (iter->proc)
 		fclose(iter->proc);
+	if (iter->sysblock)
+		closedir(iter->sysblock);
 	iter->minors = NULL;
 	iter->proc = NULL;
+	iter->sysblock = NULL;
 	iter->done = 1;
 	return 0;
 }
@@ -365,8 +403,8 @@ static int loopiter_set_device(struct loopdev_cxt *lc, const char *device)
 	if ((lc->iter.flags & LOOPITER_FL_FREE) && !used)
 		return 0;
 
-	DBG(lc, loopdev_debug("iter: setting device"));
-	loopcxt_set_device(lc, NULL);
+	DBG(lc, loopdev_debug("iter: unset device"));
+	ignore_result( loopcxt_set_device(lc, NULL) );
 	return 1;
 }
 
@@ -442,6 +480,85 @@ static int loop_scandir(const char *dirname, int **ary, int hasprefix)
 	return count;
 }
 
+/*
+ * Set the next *used* loop device according to /proc/partitions.
+ *
+ * Loop devices smaller than 512 bytes are invisible for this function.
+ */
+static int loopcxt_next_from_proc(struct loopdev_cxt *lc)
+{
+	struct loopdev_iter *iter = &lc->iter;
+	char buf[BUFSIZ];
+
+	DBG(lc, loopdev_debug("iter: scan /proc/partitions"));
+
+	if (!iter->proc)
+		iter->proc = fopen(_PATH_PROC_PARTITIONS, "r" UL_CLOEXECSTR);
+	if (!iter->proc)
+		return 1;
+
+	while (fgets(buf, sizeof(buf), iter->proc)) {
+		unsigned int m;
+		char name[128 + 1];
+
+
+		if (sscanf(buf, " %u %*s %*s %128[^\n ]",
+			   &m, name) != 2 || m != LOOPDEV_MAJOR)
+			continue;
+
+		DBG(lc, loopdev_debug("iter: check %s", name));
+
+		if (loopiter_set_device(lc, name) == 0)
+			return 0;
+	}
+
+	return 1;
+}
+
+/*
+ * Set the next *used* loop device according to
+ * /sys/block/loopN/loop/backing_file (kernel >= 2.6.37 is required).
+ *
+ * This is preferred method.
+ */
+static int loopcxt_next_from_sysfs(struct loopdev_cxt *lc)
+{
+	struct loopdev_iter *iter = &lc->iter;
+	struct dirent *d;
+	int fd;
+
+	DBG(lc, loopdev_debug("iter: scan /sys/block"));
+
+	if (!iter->sysblock)
+		iter->sysblock = opendir(_PATH_SYS_BLOCK);
+
+	if (!iter->sysblock)
+		return 1;
+
+	fd = dirfd(iter->sysblock);
+
+	while ((d = readdir(iter->sysblock))) {
+		char name[256];
+		struct stat st;
+
+		DBG(lc, loopdev_debug("iter: check %s", d->d_name));
+
+		if (strcmp(d->d_name, ".") == 0
+		    || strcmp(d->d_name, "..") == 0
+		    || strncmp(d->d_name, "loop", 4) != 0)
+			continue;
+
+		snprintf(name, sizeof(name), "%s/loop/backing_file", d->d_name);
+		if (fstat_at(fd, _PATH_SYS_BLOCK, name, &st, 0) != 0)
+			continue;
+
+		if (loopiter_set_device(lc, d->d_name) == 0)
+			return 0;
+	}
+
+	return 1;
+}
+
 /*
  * @lc: context, has to initialized by loopcxt_init_iterator()
  *
@@ -465,23 +582,14 @@ int loopcxt_next(struct loopdev_cxt *lc)
 	/* A) Look for used loop devices in /proc/partitions ("losetup -a" only)
 	 */
 	if (iter->flags & LOOPITER_FL_USED) {
-		char buf[BUFSIZ];
-
-		if (!iter->proc)
-			iter->proc = fopen(_PATH_PROC_PARTITIONS, "r");
-
-		while (iter->proc && fgets(buf, sizeof(buf), iter->proc)) {
-			unsigned int m;
-			char name[128];
-
-			if (sscanf(buf, " %u %*s %*s %128[^\n ]",
-				   &m, name) != 2 || m != LOOPDEV_MAJOR)
-				continue;
-
-			if (loopiter_set_device(lc, name) == 0)
-				return 0;
-		}
+		int rc;
 
+		if (loopcxt_sysfs_available(lc))
+			rc = loopcxt_next_from_sysfs(lc);
+		else
+			rc = loopcxt_next_from_proc(lc);
+		if (rc == 0)
+			return 0;
 		goto done;
 	}
 
@@ -489,6 +597,7 @@ int loopcxt_next(struct loopdev_cxt *lc)
 	 *    of loop devices). This is enough for 99% of all cases.
 	 */
 	if (iter->default_check) {
+		DBG(lc, loopdev_debug("iter: next: default check"));
 		for (++iter->ncur; iter->ncur < LOOPDEV_DEFAULT_NNODES;
 							iter->ncur++) {
 			char name[16];
@@ -503,6 +612,7 @@ int loopcxt_next(struct loopdev_cxt *lc)
 	/* C) the worst possibility, scan whole /dev or /dev/loop/<N>
 	 */
 	if (!iter->minors) {
+		DBG(lc, loopdev_debug("iter: next: scan /dev"));
 		iter->nminors = (lc->flags & LOOPDEV_FL_DEVSUBDIR) ?
 			loop_scandir(_PATH_DEV_LOOP, &iter->minors, 0) :
 			loop_scandir(_PATH_DEV, &iter->minors, 1);
@@ -758,7 +868,7 @@ int loopmod_supports_partscan(void)
 	if (get_linux_version() >= KERNEL_VERSION(3,2,0))
 		return 1;
 
-	f = fopen("/sys/module/loop/parameters/max_part", "r");
+	f = fopen("/sys/module/loop/parameters/max_part", "r" UL_CLOEXECSTR);
 	if (!f)
 		return 0;
 	rc = fscanf(f, "%d", &ret);
@@ -962,59 +1072,93 @@ int loopcxt_set_backing_file(struct loopdev_cxt *lc, const char *filename)
 	return 0;
 }
 
-static int digits_only(const char *s)
-{
-	while (*s)
-		if (!isdigit(*s++))
-			return 0;
-	return 1;
-}
-
 /*
- * @lc: context
- * @encryption: encryption name / type (see lopsetup man page)
- * @password
+ * In kernels prior to v3.9, if the offset or sizelimit options
+ * are used, the block device's size won't be synced automatically.
+ * blockdev --getsize64 and filesystems will use the backing
+ * file size until the block device has been re-opened or the
+ * LOOP_SET_CAPACITY ioctl is called to sync the sizes.
  *
- * Note that the encryption functionality is deprecated an unmaintained. Use
- * cryptsetup (it also supports AES-loops).
- *
- * The setting is removed by loopcxt_set_device() loopcxt_next()!
- *
- * Returns: 0 on success, <0 on error.
+ * Since mount -oloop uses the LO_FLAGS_AUTOCLEAR option and passes
+ * the open file descriptor to the mount system call, we need to use
+ * the ioctl. Calling losetup directly doesn't have this problem since
+ * it closes the device when it exits and whatever consumes the device
+ * next will re-open it, causing the resync.
  */
-int loopcxt_set_encryption(struct loopdev_cxt *lc,
-			   const char *encryption,
-			   const char *password)
+static int loopcxt_check_size(struct loopdev_cxt *lc, int file_fd)
 {
-	if (!lc)
-		return -EINVAL;
+	uint64_t size, expected_size;
+	int dev_fd;
+	struct stat st;
 
-	DBG(lc, loopdev_debug("setting encryption '%s'", encryption));
+	if (!lc->info.lo_offset && !lc->info.lo_sizelimit)
+		return 0;
 
-	if (encryption && *encryption) {
-		if (digits_only(encryption)) {
-			lc->info.lo_encrypt_type = atoi(encryption);
-		} else {
-			lc->info.lo_encrypt_type = LO_CRYPT_CRYPTOAPI;
-			snprintf((char *)lc->info.lo_crypt_name, LO_NAME_SIZE,
-				 "%s", encryption);
+	if (fstat(file_fd, &st)) {
+		DBG(lc, loopdev_debug("failed to fstat backing file"));
+		return -errno;
+	}
+	if (S_ISBLK(st.st_mode)) {
+		if (blkdev_get_size(file_fd,
+				(unsigned long long *) &expected_size)) {
+			DBG(lc, loopdev_debug("failed to determine device size"));
+			return -errno;
+		}
+	} else
+		expected_size = st.st_size;
+
+	if (expected_size == 0 || expected_size <= lc->info.lo_offset) {
+		DBG(lc, loopdev_debug("failed to determine expected size"));
+		return 0;	/* ignore this error */
+	}
+
+	if (lc->info.lo_offset > 0)
+		expected_size -= lc->info.lo_offset;
+
+	if (lc->info.lo_sizelimit > 0 && lc->info.lo_sizelimit < expected_size)
+		expected_size = lc->info.lo_sizelimit;
+
+	dev_fd = loopcxt_get_fd(lc);
+	if (dev_fd < 0) {
+		DBG(lc, loopdev_debug("failed to get loop FD"));
+		return -errno;
+	}
+
+	if (blkdev_get_size(dev_fd, (unsigned long long *) &size)) {
+		DBG(lc, loopdev_debug("failed to determine loopdev size"));
+		return -errno;
+	}
+
+	/* It's block device, so, align to 512-byte sectors */
+	if (expected_size % 512) {
+		DBG(lc, loopdev_debug("expected size misaligned to 512-byte sectors"));
+		expected_size = (expected_size >> 9) << 9;
+	}
+
+	if (expected_size != size) {
+		DBG(lc, loopdev_debug("warning: loopdev and expected "
+				      "size dismatch (%ju/%ju)",
+				      size, expected_size));
+
+		if (loopcxt_set_capacity(lc)) {
+			/* ioctl not available */
+			if (errno == ENOTTY || errno == EINVAL)
+				errno = ERANGE;
+			return -errno;
+		}
+
+		if (blkdev_get_size(dev_fd, (unsigned long long *) &size))
+			return -errno;
+
+		if (expected_size != size) {
+			errno = ERANGE;
+			DBG(lc, loopdev_debug("failed to set loopdev size, "
+					"size: %ju, expected: %ju",
+					size, expected_size));
+			return -errno;
 		}
 	}
 
-	switch (lc->info.lo_encrypt_type) {
-	case LO_CRYPT_NONE:
-		lc->info.lo_encrypt_key_size = 0;
-		break;
-	default:
-		DBG(lc, loopdev_debug("setting encryption key"));
-		memset(lc->info.lo_encrypt_key, 0, LO_KEY_SIZE);
-		strncpy((char *)lc->info.lo_encrypt_key, password, LO_KEY_SIZE);
-		lc->info.lo_encrypt_key[LO_KEY_SIZE - 1] = '\0';
-		lc->info.lo_encrypt_key_size = LO_KEY_SIZE;
-		break;
-	}
-
-	DBG(lc, loopdev_debug("encryption successfully set"));
 	return 0;
 }
 
@@ -1051,7 +1195,7 @@ int loopcxt_setup_device(struct loopdev_cxt *lc)
 	if (lc->info.lo_flags & LO_FLAGS_READ_ONLY)
 		mode = O_RDONLY;
 
-	if ((file_fd = open(lc->filename, mode)) < 0) {
+	if ((file_fd = open(lc->filename, mode | O_CLOEXEC)) < 0) {
 		if (mode != O_RDONLY && (errno == EROFS || errno == EACCES))
 			file_fd = open(lc->filename, mode = O_RDONLY);
 
@@ -1063,6 +1207,7 @@ int loopcxt_setup_device(struct loopdev_cxt *lc)
 	DBG(lc, loopdev_debug("setup: backing file open: OK"));
 
 	if (lc->fd != -1 && lc->mode != mode) {
+		DBG(lc, loopdev_debug("closing already open device (mode mismatch)"));
 		close(lc->fd);
 		lc->fd = -1;
 		lc->mode = 0;
@@ -1096,9 +1241,6 @@ int loopcxt_setup_device(struct loopdev_cxt *lc)
 
 	DBG(lc, loopdev_debug("setup: LOOP_SET_FD: OK"));
 
-	close(file_fd);
-	file_fd = -1;
-
 	if (ioctl(dev_fd, LOOP_SET_STATUS64, &lc->info)) {
 		DBG(lc, loopdev_debug("LOOP_SET_STATUS64 failed: %m"));
 		goto err;
@@ -1106,6 +1248,11 @@ int loopcxt_setup_device(struct loopdev_cxt *lc)
 
 	DBG(lc, loopdev_debug("setup: LOOP_SET_STATUS64: OK"));
 
+	if ((rc = loopcxt_check_size(lc, file_fd)))
+		goto err;
+
+	close(file_fd);
+
 	memset(&lc->info, 0, sizeof(lc->info));
 	lc->has_info = 0;
 	lc->info_failed = 0;
@@ -1122,6 +1269,24 @@ err:
 	return rc;
 }
 
+int loopcxt_set_capacity(struct loopdev_cxt *lc)
+{
+	int fd = loopcxt_get_fd(lc);
+
+	if (fd < 0)
+		return -EINVAL;
+
+	/* Kernels prior to v2.6.30 don't support this ioctl */
+	if (ioctl(fd, LOOP_SET_CAPACITY, 0) < 0) {
+		int rc = -errno;
+		DBG(lc, loopdev_debug("LOOP_SET_CAPACITY failed: %m"));
+		return rc;
+	}
+
+	DBG(lc, loopdev_debug("capacity set"));
+	return 0;
+}
+
 int loopcxt_delete_device(struct loopdev_cxt *lc)
 {
 	int fd = loopcxt_get_fd(lc);
@@ -1138,6 +1303,36 @@ int loopcxt_delete_device(struct loopdev_cxt *lc)
 	return 0;
 }
 
+int loopcxt_add_device(struct loopdev_cxt *lc)
+{
+	int rc = -EINVAL;
+	int ctl, nr = -1;
+	const char *p, *dev = loopcxt_get_device(lc);
+
+	if (!dev)
+		goto done;
+
+	if (!(lc->flags & LOOPDEV_FL_CONTROL)) {
+		rc = -ENOSYS;
+		goto done;
+	}
+
+	p = strrchr(dev, '/');
+	if (!p || (sscanf(p, "/loop%d", &nr) != 1 && sscanf(p, "/%d", &nr) != 1)
+	       || nr < 0)
+		goto done;
+
+	ctl = open(_PATH_DEV_LOOPCTL, O_RDWR|O_CLOEXEC);
+	if (ctl >= 0) {
+		DBG(lc, loopdev_debug("add_device %d", nr));
+		rc = ioctl(ctl, LOOP_CTL_ADD, nr);
+		close(ctl);
+	}
+done:
+	DBG(lc, loopdev_debug("add_device done [rc=%d]", rc));
+	return rc;
+}
+
 /*
  * Note that LOOP_CTL_GET_FREE ioctl is supported since kernel 3.1. In older
  * kernels we have to check all loop devices to found unused one.
@@ -1151,7 +1346,7 @@ int loopcxt_find_unused(struct loopdev_cxt *lc)
 	DBG(lc, loopdev_debug("find_unused requested"));
 
 	if (lc->flags & LOOPDEV_FL_CONTROL) {
-		int ctl = open(_PATH_DEV_LOOPCTL, O_RDWR);
+		int ctl = open(_PATH_DEV_LOOPCTL, O_RDWR|O_CLOEXEC);
 
 		if (ctl >= 0)
 			rc = ioctl(ctl, LOOP_CTL_GET_FREE);
@@ -1191,27 +1386,29 @@ int loopdev_is_autoclear(const char *device)
 	if (!device)
 		return 0;
 
-	loopcxt_init(&lc, 0);
-	loopcxt_set_device(&lc, device);
-	rc = loopcxt_is_autoclear(&lc);
-	loopcxt_deinit(&lc);
+	rc = loopcxt_init(&lc, 0);
+	if (!rc)
+		rc = loopcxt_set_device(&lc, device);
+	if (!rc)
+		rc = loopcxt_is_autoclear(&lc);
 
+	loopcxt_deinit(&lc);
 	return rc;
 }
 
 char *loopdev_get_backing_file(const char *device)
 {
 	struct loopdev_cxt lc;
-	char *res;
+	char *res = NULL;
 
 	if (!device)
 		return NULL;
+	if (loopcxt_init(&lc, 0))
+		return NULL;
+	if (loopcxt_set_device(&lc, device) == 0)
+		res = loopcxt_get_backing_file(&lc);
 
-	loopcxt_init(&lc, 0);
-	loopcxt_set_device(&lc, device);
-	res = loopcxt_get_backing_file(&lc);
 	loopcxt_deinit(&lc);
-
 	return res;
 }
 
@@ -1228,8 +1425,11 @@ int loopdev_is_used(const char *device, const char *filename,
 	if (!device || !filename)
 		return 0;
 
-	loopcxt_init(&lc, 0);
-	loopcxt_set_device(&lc, device);
+	rc = loopcxt_init(&lc, 0);
+	if (!rc)
+		rc = loopcxt_set_device(&lc, device);
+	if (rc)
+		return rc;
 
 	rc = !stat(filename, &st);
 	rc = loopcxt_is_used(&lc, rc ? &st : NULL, filename, offset, flags);
@@ -1246,8 +1446,9 @@ int loopdev_delete(const char *device)
 	if (!device)
 		return -EINVAL;
 
-	loopcxt_init(&lc, 0);
-	rc = loopcxt_set_device(&lc, device);
+	rc = loopcxt_init(&lc, 0);
+	if (!rc)
+		rc = loopcxt_set_device(&lc, device);
 	if (!rc)
 		rc = loopcxt_delete_device(&lc);
 	loopcxt_deinit(&lc);
@@ -1294,8 +1495,9 @@ char *loopdev_find_by_backing_file(const char *filename, uint64_t offset, int fl
 	if (!filename)
 		return NULL;
 
-	loopcxt_init(&lc, 0);
-	if (loopcxt_find_by_backing_file(&lc, filename, offset, flags))
+	if (loopcxt_init(&lc, 0))
+		return NULL;
+	if (loopcxt_find_by_backing_file(&lc, filename, offset, flags) == 0)
 		res = loopcxt_strdup_device(&lc);
 	loopcxt_deinit(&lc);
 
@@ -1310,12 +1512,14 @@ char *loopdev_find_by_backing_file(const char *filename, uint64_t offset, int fl
 int loopdev_count_by_backing_file(const char *filename, char **loopdev)
 {
 	struct loopdev_cxt lc;
-	int count = 0;
+	int count = 0, rc;
 
 	if (!filename)
 		return -1;
 
-	loopcxt_init(&lc, 0);
+	rc = loopcxt_init(&lc, 0);
+	if (rc)
+		return rc;
 	if (loopcxt_init_iterator(&lc, LOOPITER_FL_USED))
 		return -1;
 
@@ -1353,7 +1557,8 @@ static void test_loop_info(const char *device, int flags, int debug)
 	char *p;
 	uint64_t u64;
 
-	loopcxt_init(&lc, flags);
+	if (loopcxt_init(&lc, flags))
+		return;
 	loopcxt_enable_debug(&lc, debug);
 
 	if (loopcxt_set_device(&lc, device))
@@ -1379,7 +1584,8 @@ static void test_loop_scan(int flags, int debug)
 	struct loopdev_cxt lc;
 	int rc;
 
-	loopcxt_init(&lc, 0);
+	if (loopcxt_init(&lc, 0))
+		return;
 	loopcxt_enable_debug(&lc, debug);
 
 	if (loopcxt_init_iterator(&lc, flags))
@@ -1405,9 +1611,11 @@ static void test_loop_scan(int flags, int debug)
 static int test_loop_setup(const char *filename, const char *device, int debug)
 {
 	struct loopdev_cxt lc;
-	int rc = 0;
+	int rc;
 
-	loopcxt_init(&lc, 0);
+	rc = loopcxt_init(&lc, 0);
+	if (rc)
+		return rc;
 	loopcxt_enable_debug(&lc, debug);
 
 	if (device) {
