@@ -20,7 +20,6 @@
 
 #include "partitions.h"
 #include "crc32.h"
-#include "dos.h"
 
 #define GPT_PRIMARY_LBA	1
 
@@ -48,7 +47,7 @@ typedef struct {
 struct gpt_header {
 	uint64_t	signature;		/* "EFI PART" */
 	uint32_t	revision;
-	uint32_t	header_size;		/* usualy 92 bytes */
+	uint32_t	header_size;		/* usually 92 bytes */
 	uint32_t	header_crc32;		/* checksum of header with this
 						 * field zeroed during calculation */
 	uint32_t	reserved1;
@@ -120,7 +119,7 @@ static inline int guidcmp(efi_guid_t left, efi_guid_t right)
 }
 
 /*
- * UUID is traditionaly 16 byte big-endian array, except Intel EFI
+ * UUID is traditionally 16 byte big-endian array, except Intel EFI
  * specification where the UUID is a structure of little-endian fields.
  */
 static void swap_efi_guid(efi_guid_t *uid)
@@ -156,13 +155,15 @@ static int last_lba(blkid_probe pr, uint64_t *lba)
  * Note that the PMBR detection is optional (enabled by default) and could be
  * disabled by BLKID_PARTS_FOPCE_GPT flag (see also blkid_paertitions_set_flags()).
  */
-static int is_pmbr_valid(blkid_probe pr)
+static int is_pmbr_valid(blkid_probe pr, int *has)
 {
 	int flags = blkid_partitions_get_flags(pr);
 	unsigned char *data;
 	struct dos_partition *p;
 	int i;
 
+	if (has)
+		*has = 0;
 	if (flags & BLKID_PARTS_FORCE_GPT)
 		goto ok;			/* skip PMBR check */
 
@@ -170,18 +171,18 @@ static int is_pmbr_valid(blkid_probe pr)
 	if (!data)
 		goto failed;
 
-	if (!is_valid_mbr_signature(data))
+	if (!mbr_is_valid_magic(data))
 		goto failed;
 
-	p = (struct dos_partition *) (data + BLKID_MSDOS_PT_OFFSET);
-
-	for (i = 0; i < 4; i++, p++) {
-		if (p->sys_type == BLKID_GPT_PARTITION)
+	for (i = 0, p = mbr_get_partition(data, 0); i < 4; i++, p++) {
+		if (p->sys_ind == MBR_GPT_PARTITION)
 			goto ok;
 	}
 failed:
 	return 0;
 ok:
+	if (has)
+		*has = 1;
 	return 1;
 }
 
@@ -226,20 +227,20 @@ static struct gpt_header *get_gpt_header(
 		return NULL;
 
 	/* Header has to be verified when header_crc32 is zero */
-	orgcrc = le32_to_cpu(h->header_crc32);
+	orgcrc = h->header_crc32;
 	h->header_crc32 = 0;
-
 	crc = count_crc32((unsigned char *) h, hsz);
-	if (crc != orgcrc) {
-		DBG(DEBUG_LOWPROBE, printf("GPT header corrupted\n"));
+	h->header_crc32 = orgcrc;
+
+	if (crc != le32_to_cpu(orgcrc)) {
+		DBG(LOWPROBE, blkid_debug("GPT header corrupted"));
 		return NULL;
 	}
-	h->header_crc32 = cpu_to_le32(orgcrc);
 
 	/* Valid header has to be at MyLBA */
 	if (le64_to_cpu(h->my_lba) != lba) {
-		DBG(DEBUG_LOWPROBE, printf(
-			"GPT->MyLBA mismatch with real position\n"));
+		DBG(LOWPROBE, blkid_debug(
+			"GPT->MyLBA mismatch with real position"));
 		return NULL;
 	}
 
@@ -248,14 +249,14 @@ static struct gpt_header *get_gpt_header(
 
 	/* Check if First and Last usable LBA makes sense */
 	if (lu < fu || fu > lastlba || lu > lastlba) {
-		DBG(DEBUG_LOWPROBE, printf(
-			"GPT->{First,Last}UsableLBA out of range\n"));
+		DBG(LOWPROBE, blkid_debug(
+			"GPT->{First,Last}UsableLBA out of range"));
 		return NULL;
 	}
 
 	/* The header has to be outside usable range */
 	if (fu < lba && lba < lu) {
-		DBG(DEBUG_LOWPROBE, printf("GPT header is inside usable area\n"));
+		DBG(LOWPROBE, blkid_debug("GPT header is inside usable area"));
 		return NULL;
 	}
 
@@ -263,7 +264,7 @@ static struct gpt_header *get_gpt_header(
 	esz = le32_to_cpu(h->num_partition_entries) *
 			le32_to_cpu(h->sizeof_partition_entry);
 	if (!esz) {
-		DBG(DEBUG_LOWPROBE, printf("GPT entries undefined\n"));
+		DBG(LOWPROBE, blkid_debug("GPT entries undefined"));
 		return NULL;
 	}
 
@@ -276,14 +277,14 @@ static struct gpt_header *get_gpt_header(
 	*ents = (struct gpt_entry *) get_lba_buffer(pr,
 				le64_to_cpu(h->partition_entries_lba), esz);
 	if (!*ents) {
-		DBG(DEBUG_LOWPROBE, printf("GPT entries unreadable\n"));
+		DBG(LOWPROBE, blkid_debug("GPT entries unreadable"));
 		return NULL;
 	}
 
 	/* Validate entries */
 	crc = count_crc32((unsigned char *) *ents, esz);
 	if (crc != le32_to_cpu(h->partition_entry_array_crc32)) {
-		DBG(DEBUG_LOWPROBE, printf("GPT entries corrupted\n"));
+		DBG(LOWPROBE, blkid_debug("GPT entries corrupted"));
 		return NULL;
 	}
 
@@ -300,12 +301,12 @@ static int probe_gpt_pt(blkid_probe pr,
 	blkid_partlist ls;
 	uint64_t fu, lu;
 	uint32_t ssf, i;
-
+	efi_guid_t guid;
 
 	if (last_lba(pr, &lastlba))
 		goto nothing;
 
-	if (!is_pmbr_valid(pr))
+	if (!is_pmbr_valid(pr, NULL))
 		goto nothing;
 
 	h = get_gpt_header(pr, &hdr, &e, (lba = GPT_PRIMARY_LBA), lastlba);
@@ -322,9 +323,15 @@ static int probe_gpt_pt(blkid_probe pr,
 			      (unsigned char *) GPT_HEADER_SIGNATURE_STR))
 		goto err;
 
-	if (blkid_partitions_need_typeonly(pr))
-		/* caller does not ask for details about partitions */
+	guid = h->disk_guid;
+	swap_efi_guid(&guid);
+
+	if (blkid_partitions_need_typeonly(pr)) {
+		/* Non-binary interface -- caller does not ask for details
+		 * about partitions, just set generic varibles only. */
+		blkid_partitions_set_ptuuid(pr, (unsigned char *) &guid);
 		return 0;
+	}
 
 	ls = blkid_probe_get_partlist(pr);
 	if (!ls)
@@ -334,6 +341,8 @@ static int probe_gpt_pt(blkid_probe pr,
 	if (!tab)
 		goto err;
 
+	blkid_parttable_set_uuid(tab, (const unsigned char *) &guid);
+
 	ssf = blkid_probe_get_sectorsize(pr) / 512;
 
 	fu = le64_to_cpu(h->first_usable_lba);
@@ -353,8 +362,8 @@ static int probe_gpt_pt(blkid_probe pr,
 		}
 		/* the partition has to inside usable range */
 		if (start < fu || start + size - 1 > lu) {
-			DBG(DEBUG_LOWPROBE, printf(
-				"GPT entry[%d] overflows usable area - ignore\n",
+			DBG(LOWPROBE, blkid_debug(
+				"GPT entry[%d] overflows usable area - ignore",
 				i));
 			blkid_partlist_increment_partno(ls);
 			continue;
@@ -369,14 +378,13 @@ static int probe_gpt_pt(blkid_probe pr,
 			(unsigned char *) e->partition_name,
 			sizeof(e->partition_name), BLKID_ENC_UTF16LE);
 
-		swap_efi_guid(&e->unique_partition_guid);
-		swap_efi_guid(&e->partition_type_guid);
+		guid = e->unique_partition_guid;
+		swap_efi_guid(&guid);
+		blkid_partition_set_uuid(par, (const unsigned char *) &guid);
 
-		blkid_partition_set_uuid(par,
-			(const unsigned char *) &e->unique_partition_guid);
-
-		blkid_partition_set_type_uuid(par,
-			(const unsigned char *) &e->partition_type_guid);
+		guid = e->partition_type_guid;
+		swap_efi_guid(&guid);
+		blkid_partition_set_type_uuid(par, (const unsigned char *) &guid);
 
 		blkid_partition_set_flags(par, e->attributes);
 	}
@@ -407,3 +415,39 @@ const struct blkid_idinfo gpt_pt_idinfo =
 	.magics		= BLKID_NONE_MAGIC
 };
 
+
+
+/* probe for *alone* protective MBR */
+static int probe_pmbr_pt(blkid_probe pr,
+		const struct blkid_idmag *mag __attribute__((__unused__)))
+{
+	int has = 0;
+	struct gpt_entry *e;
+	uint64_t lastlba = 0;
+	struct gpt_header hdr;
+
+	if (last_lba(pr, &lastlba))
+		goto nothing;
+
+	is_pmbr_valid(pr, &has);
+	if (!has)
+		goto nothing;
+
+	if (!get_gpt_header(pr, &hdr, &e, GPT_PRIMARY_LBA, lastlba) &&
+	    !get_gpt_header(pr, &hdr, &e, lastlba, lastlba))
+		return 0;
+nothing:
+	return 1;
+}
+
+const struct blkid_idinfo pmbr_pt_idinfo =
+{
+	.name		= "PMBR",
+	.probefunc	= probe_pmbr_pt,
+	.magics		=
+	{
+		{ .magic = "\x55\xAA", .len = 2, .sboff = 510 },
+		{ NULL }
+	}
+};
+
